import time
from math import floor

from epics import caget, caput

from iexcode.instruments.IEX_endstations import BL
from iexcode.instruments.scalers import scaler_cts


class Motors:
    """ 
    short hand to move motors in endstation
    usage motors = Motor('ARPES','ARPES_motor_dictionary')
    """

    def __init__(self,endstation_name,motor_dictionary,physical_motors,
    psuedo_motors):
        self.endstation_name = endstation_name
        self._motor_dictionary = motor_dictionary
        self.physical_motors = physical_motors
        self.pseudo_motors = psuedo_motors
           
    def info(self):
        print("physical motors:",self.physical_motors)
        print("pseudo motors:",self.pseudo_motors)
    
    def get(self,name,verbose=False):   
        """
        get current position of motor name = "x","y"...
        """   

        try:
            rbv = caget(self._motor_dictionary[name][0])
            if verbose:
                print('current position: '+name+" = "+str(rbv))
            return rbv
        except:
            print("Not a valid motor name") 
    
    def reset(self,name):
        """
        Reset motor if stuck in 'moving'
        """
        pv=self._motor_dictionary[name][2]
        caput(pv,'Stop')
        time.sleep(1)
        caput(pv,'Go')
        
    def status(self):
        """
        if status == 0 then moving
        """
        status = 0
        for motor in self._motor_dictionary.keys():
            status *= caget(self._motor_dictionary[motor][3]+".DMOV")
        return status

    def move(self,name,val,**kwargs):
        """
        Moves a motor in the endstation based on common name, not PV name
        
        name: motor name => 'x','y','z','th','chi','phi' ...
        val: absolute position if relative=False, delta if relative=True
        
        **kwargs:
            relative: move mode => True/False; default = False (absolute)
            wait: wait for move completion before moving on => True/False; default=True
            verbose: print old and new motor position => True/False; default=False


        Previously: Move_ARPES_Motor or Move_ARPES_Motor or motor_move_vs_Branch
        """
        kwargs.setdefault('relative',False)
        kwargs.setdefault('wait',True)
        kwargs.setdefault('verbose',False)

        try:
            rbv_pv,val_pv,sgm_pv,pv = self._motor_dictionary[name]
            old_position = caget(rbv_pv)
            if kwargs['relative']:
                val = old_position + val
            
            caput(val_pv,val,wait=kwargs['wait'],timeout=18000)
            
            new_position = caget(rbv_pv)
            if kwargs['verbose']:
                print('old: '+name+" = "+str(old_position))
                print('new: '+name+" = "+str(new_position))
        except:
            print(name+" is not a valid motor name")



    def dmove(self,name,val,**kwargs):
        """
        relative move of a motor in the endstation based on common name, not PV name
        name = x,y,z,th,chi,phi ...

        **kwargs:
            wait: wait for move completion before moving on => True/False; default=True
            verbose: print old and new motor position => True/False; default=False

        Previously: motor_umove_vs_Branch
        """
        kwargs.setdefault('wait',True)
        kwargs.setdefault('verbose',False)

        self.move(name,val,relative=True,wait=kwargs['wait'],verbose=kwargs['verbose'])

    def mprint(self):
        """
        prints current position of the physical motors
        """
        position_list = []
        for motor in self.physical_motors:
            position_list.append(self.get(motor,verbose=False))
        return position_list

    def mvsample(self,position_list,verbose=True):
        """
        moves the sample to the position sepcified by position_list
        position_list = ['description',x,y,z,th,chi,phi]

        Previously: sample
        """
        if not isinstance(position_list[0],str):
            sample_name = ''
        else:
            sample_name = position_list[0]
            position_list = position_list[1:]

        motor_list = self.physical_motors
        #debug
        print('motor_list=',motor_list)
        print('position_list=',position_list)
        for i,motor in enumerate(motor_list):
            self.move(motor,position_list[i])

        if verbose:
            print("Sample now @ "+sample_name)

    def scan(self,name,start,stop,step, **kwargs):
        """
        scans a motor
        
        **kwargs:
            relative: move mode => True/False; default = False (absolute)
            scan_dim: 1  (default)
            execute: True/False to start the scan => True (default)

            for Kappa only:
                cts: scaler integration time in seconds (default = 0.1 s)
                mpa: True/False to sum mpa acquisitions
            
            Previously Scan_ARPES_motor / Scan_Kappa_motor
        """
        kwargs.setdefault('relative',False)
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault('execute',True)
        #for kappa only
        kwargs.setdefault('cts',0.1)
        kwargs.setdefault('mpa',False)

        rbv_pv,val_pv,sgm_pv,pv =self._motor_dictionary()[name]

        if kwargs['mode'] == "relative":
            current_value = caget(rbv_pv)
            abs_start = round(current_value + start,3)
            abs_stop  = round(current_value + stop,3)
            print("start, stop, step = "+str(abs_start)+", "+str(abs_stop)
            +", "+str(step))
        else:
            abs_start = start
            abs_stop = stop
        
        self._scalar_cts(kwargs['cts'],verbose=True,**kwargs)
        BL.mda.fillin(val_pv,rbv_pv,abs_start,abs_stop,step,**kwargs)
        
        if kwargs['execute']:
            BL.mda.go(**kwargs)


    def scan_2D(self,inner_loop_list,outer_loop_list,**kwargs):
        """
        Fills in a the Scan Record for a 2D scan (Mesh), does NOT press go
        InnerMotorList=[name1,start1,stop1,step1] #scanDIM=2
        OuterMotorList=[name2,start2,stop2,step2] #scanDIM=3
            name = 'x'/'y'/'z'/'th'...    
        
        **kwargs:
            relative: scab => True/False; default = False (absolute)
            outer_scan_dim: 2  (default)
            execute: True/False to start the scan => True (default)
            Snake; coming soon


            for Kappa only:
                cts: scaler integration time in seconds (default = 0.1 s)
                mpa: True/False to sum mpa acquisitions

        """
        kwargs.setdefault('mode','absolute')
        kwargs.setdefault('outer_scan_dim',2)
        kwargs.setdefault('snake',False)
        kwargs.setdefault('execute',True)
        #for kappa only
        kwargs.setdefault('cts',0.1)
        kwargs.setdefault('mpa',False)

        rbv_pv,val_pv,sgm_pv,pv = self._motor_dictionary()[inner_loop_list[0]]
        inner_loop_list.insert(0,rbv_pv)
        inner_loop_list.insert(0,val_pv)
        
        rbv_pv,val_pv,sgm_pv,pv = self._motor_dictionary()[outer_loop_list[0]]
        outer_loop_list.insert(0,rbv_pv)
        outer_loop_list.insert(0,val_pv)

        if kwargs["mode"] == "relative":
            current_value1 = caget(inner_loop_list[1])
            inner_loop_list[2]=round(current_value1+inner_loop_list[2],3)
            inner_loop_list[3]=round(current_value1+inner_loop_list[3],3)

            current_value2 = caget(outer_loop_list[1])
            outer_loop_list[2]=round(current_value1+outer_loop_list[2],3)
            outer_loop_list[3]=round(current_value1+outer_loop_list[3],3)

        scaler_cts(kwargs['cts'],verbose=True,**kwargs)
        BL.mda.fillin_2D(inner_loop_list,outer_loop_list,
        outer_scan_dim=kwargs['outer_scan_dim'],**kwargs)

        if kwargs['execute']:
            BL.mda.go(**kwargs)




