from os.path import join,isfile

from epics import caget, caput
from iexcode.instruments.IEX_endstations import *
from iexcode.instruments.utilities import today


##############################################################################################################
##############################             logging           ##############################
##############################################################################################################
def log_print(**kwargs):
    """
    prints a comment to the logfile
    """
    logfile_print(BL.endstation_name,BL.ioc,kwargs['comment'])

def log_update():
    """
    updates the log file with the last scan info
    """
    try:
        entry_list,pv_list, format_list = BL.log_entries()
        logfile_update(BL.endstation_name,BL.ioc,entry_list,pv_list,format_list)
    except:
        print('No logfile written')


##############################################################################################################
##############################            general   logfile functions          ##############################
##############################################################################################################


def logfile_name_pv(endstation_name):
    """
    Dictionary to get the PV in which the FileName for logging is stored
    
    endstation: 'ARPES' / 'Kappa' / 'Test

    Previously: logname_pv
    """
    pv_dict={
        'Test':'29idb:userStringSeq10.STR1',
        'ARPES':'29idb:userStringSeq10.STR2',
        'Kappa':'29idb:userStringSeq10.STR3',
        'RSoXS':'29idb:userStringSeq10.STR4',
        }
    if endstation_name in pv_dict.keys():
        pv=pv_dict[endstation_name]
    else:
        pv='29idb:userStringSeq10.STR10'
    return pv

def logfile_name_set(endstation_name,**kwargs):
    """
    Sets the string used for the FileName in scanlog and EAlog
    uses logname_PV to reference PV associated with a particular ioc
    **kwargs:
        filename = to set a specific filename, default => YYYYMMDD_log.txt

    Previously: logname_set
    """
    kwargs.setdefault('filename_suffix','_log')
    kwargs.setdefault('filename',today()+kwargs['filename_suffix']+'.txt')

    filename = kwargs['filename']

    try:
        caput(logfile_name_pv(endstation_name),filename)
        print("\nLog filename = \'"+filename+"\' @ "+logfile_name_pv(endstation_name))
        print('To change filename, use logname_set(endstation_name,new_name)')
    except:
        print("Error: was not able to set the filename, check that 29idb ioc is running")

        
def logfile_name_get(endstation_name):
    """
    Gets the string used for the FileName in scanlog and EAlog for the given scanIOC.

    Previously: logname
    """

    return caget(logfile_name_pv(endstation_name))
       
def logfile_fpath(endstation_name,mda_ioc):
    """
    returns the full path to the logfile based on the current user in the mda scanRecord


    Previously: logname, logname_generate
    """
 
    try: 
        filename = logfile_name_get(endstation_name)
        user_name = BL.mda.scanRecord_user(mda_ioc)
        fpath_with_subfolder = join(user_name,filename)
    except:
        fpath_with_subfolder = "logfile.txt"
        print("Couldn't read log file path, using: "+fpath_with_subfolder)
        
    return fpath_with_subfolder


def logfile_print(endstation_name,mda_ioc,comment=''):
    """
    Writes comments to the logfile on a new line
    comment is a string of what you want to add to the file 
        e.g. comment = "------- New Sample--------"
    
    Previously: logprint
    """

    logfile_name = logfile_name_get(endstation_name)
    
    try:
        fpath_with_subfolder=logfile_fpath(endstation_name,mda_ioc) 
        if not isfile(fpath_with_subfolder):
            logfile_header(fpath_with_subfolder,entry=None)
        with open(fpath_with_subfolder,'a') as myfile:
            myfile.write(comment)
            myfile.write('\n')
        #print(comment,FileNameWithSubFolder)
    except:
        print("Logprint failed")


def logfile_header(endstation_name,mda_ioc,header_list): 
    """
    file path = path to where file_name lives
    file_name = name of the logfile
    entry is a list

    Previously: SaveFile_Header
    """  
    version = '1.4'

    fpath_with_subfolder=logfile_fpath(endstation_name,mda_ioc) 
    with open(fpath_with_subfolder, "w+") as f:
        f.write('@Log version: '+ version+'\n\n')
        file_path  = BL.mda.scanRecord_filepath(mda_ioc)
        f.write('FilePath '+endstation_name+': '+file_path+'\n')

        for key in header_list.keys():
            f.write(key+' Header:  '+ header_list[key]+'\n\n')


   
def logfile_update(endstation_name,mda_ioc,entry_list,pv_list,format_list):
    """ 
    To be used for scanlog and scanEA functions.
    Update SaveFile_Header version number when changing the structure of the file (indexing).

    Previously: SaveFile
    """
    fpath_with_subfolder=logfile_fpath(endstation_name,mda_ioc) 
    
    if not isfile(fpath_with_subfolder):
        logfile_header(endstation_name,mda_ioc,entry_list)

    with open(fpath_with_subfolder, "a+") as f:
        for i in range(len(format_list)):
            pv_format = "{0:"+format_list[i]+"},"
            f.write(pv_format.format(pv_list[i]))
        last_entry=len(format_list)-1
        pv_format="{0:"+format_list[last_entry]+"}\n"
        f.write(pv_format.format(pv_list[last_entry]))


