import matplotlib.pyplot as plt

from ..instruments.diagnostics import diodeC
from ..instruments.files_and_folders import check_run
from ..instruments.utilities import wait_for_it
from ..instruments.current_amplifiers import Keithley
from ..instruments.IEX_VPU import *
from ..instruments.VLS_PGM import *
from ..instruments.xrays import *
from ..instruments.shutters import *
from ..instruments.slits import *

from ..instruments.Kappa import *
from ..instruments.ARPES import *

from .commissioning import *
from .ScanFunctions_plot import *

def start_of_the_week(grt,branch,wait=False,**kwargs):
    """
    This should be run every at the start of every week; 50 min total.
    Switch to C-branch,.
    If wait=True, wait for next day a 8:05.
    
    kwargs defaults:
        run = None: used today's date to set run appropriatetly 
        repeat = False: sets everything: switches branch, sets BL_Mode to Staff, resets scan/CAs/mono
        scanType = ['slit1','wire','flux','monoVslit']
        sound = False; no sounds
              = True => plays sound after Slit1A scans and when script is complete
    }
    ID steering: (slit1A)
        steering out => move beam more positive (10 urad ~ 0.25 mm)
        steering up  => move beam more positive (10 urad ~ 0.25 mm)
        
    M1 steering: (monoVslit-2V and 2H)
        Roll:  Rz more positive => 2V more positive (0.25 => .25 mm)
        Pitch: Ry more positive => 2H more positive (0.03 => .5 mm

    Previously: StartOfTheWeek
    """
    kwargs.setdefault('repeat',False)
    kwargs.setdefault('run',check_run())
    kwargs.setdefault('quick',True)    
    kwargs.setdefault('interactive',True)    
    kwargs.setdefault('scanType',['slit1','wire','flux','monoVslit'])
    kwargs.setdefault('sound',True)
    kwargs.setdefault('extended_range',False)
    
    for scan in kwargs['scanType']: 
        if scan not in ['slit1','wire','flux','monoVslit']:
            print(scan+" is not a valid scan scanType=['slit1','wire','flux','monoVslit']")
            return

        
    ### checking diode settings:
    if branch == 'c': 
        diode = Keithley('b',15)
        diode.autoscale()
        diodeC('In')
    else:
        foo=input('Do you have a diode in direct beam (y or n)? >')
        if foo.lower() == 'y' or foo.lower() == 'yes':
            print('Resuming...')
    
    
    ### checking QP ratio:
    QP_ratio = ID_QP_ratio_get()
    if QP_ratio != 100:
        foo=input('QP on!!! Continue (y or n)? >')
        if foo.lower() == 'y' or foo.lower() == 'yes':
            print('Resuming...')
        
    ### Stuff that doesn't need beam & does not need to be run if repeating SoTW:
    print("\n\n================== Sets Beamline & Scans:")
    switch_gratings(grt)

    if not kwargs['repeat']: 
        switch_branch(branch)

        print("\n\n================== Sets Directory:")

    if branch=='c':
        ARPES_init(BL_mode='staff')

    fname='StartOfTheWeek_log.txt'
    logfile_name_set(BL.endstation,filename=fname)

        
    ### checking branch shutter:
    if branch_shutter_status() == False:
        branch_shutter_open()
    

    ### Wait for next 8AM:
    if wait:
        t = datetime.today()
        if 0 <= t.hour <= 8:
            wait_for_it(0,8,5)
        else:    
            wait_for_it(1,8,5)

    ### checking ID:
    print("\n\n================== Start ID:")
    polarization('RCP')
    
        
    ### Ready to start:
    FirstScan=BL.mda.fileNum+1


    if not kwargs['repeat']:    
        log_print("\n\n================== Start Of The Week @ "+today('slash')+':\n')
        
        
    def interactive_fct(comment='scans'): 
        foo=input('\nRepeat '+comment+' (y or n)? >')
        if foo.lower() == 'y' or foo.lower() ==  'yes': 
            print('\nRepeating...\n');flag=True
            
        else: 
            print('\nResuming...\n'); flag=False
        return flag
    
    DetDict={'c':(9,7,8,15),'d':(9,7,8,14)}
    detCA4,detH,detV,detDiode=DetDict[branch]

    ###### Scan front-end slit center:
    if 'slit1' in kwargs['scanType']:
        print("\n\n================== Slit 1A scans:")
        ID_SP_set(2000)
        flag=True
        while flag:
            
            #Scan_SlitCenter('1H',-3,3,0.1,comment='Slit center - 1H')
            #Scan_SlitCenter('1V',-3,3,0.1,comment='Slit center - 1V')
            scan_narrow_slit('1H',[0.25,-3,3,0.1],comment='Slit center - 1H')
            scan_narrow_slit('1V',[0.25,-3,3,0.1],comment='Slit center - 1V')
            #SetSlit1A(4.5,4.5,0,0)
            slit1A_set((4.5,4.5),(0,0))
            m=BL.mda.lastFileNum()
            print('\nsteering out => move beam more positive (10 urad ~ 0.25 mm)')
            print('steering up  => move beam more positive (10 urad ~ 0.25 mm)')
            try:
                h_position=fit_mda(m-1,detCA4,1,'gauss',xrange=[-1,1]);plt.show()
                v_position=fit_mda(m  ,detCA4,1,'gauss',xrange=[-1,1]);plt.show()
                print('\nBeam position:')
                print('   - horizontal: '+str(h_position))
                print('   - vertical:   '+str(v_position))
                h_steering,v_steering=beamsteering_1A(h_position,v_position)
                print('\nBeam steering:')
                print('   - horizontal: '+h_steering)
                print('   - vertical:   '+v_steering)
                print('\n')       
            except:
                print('Unable to fit position; try to tweak xrange or FWHM:')
                print('fit_mda('+str(m)+','+str(detCA4)+',1,"gauss",xrange=[-1,1])')
                print('fit_mda('+str(m+1)+','+str(detCA4)+',1,"gauss",xrange=[-1,1])')
                print('\n')
            if kwargs['sound']:playsound()
            if kwargs['interactive']: flag=interactive_fct()
            else: break

    ###### Wire scans: 
    if 'wire' in kwargs['scanType']:
        ID_SP_set(2000)
        slit1A_set(4.5,4.5,0,0)
        scan_wire('H',comment='Wire scan - H',all_diag=False)     # by default puts all meshes in + DiodeC
        scan_wire('V',comment='Wire scan - V',all_diag=False)
        m=BL.mda.lastFileNum()
        try:
            plot_mda(m-1,detH,m,detV,title='wire-H (blue) & wire-V (orange)');plt.show()
        except:
            print('Unable to plot.')
        if kwargs['sound']:playsound()
        

    ###### Mono/slit scans: 
    list_position=[]
    if 'monoVslit' in kwargs['scanType']:      
        if grt=="HEG":
            slit(300)
            c=2
        else:
            slit(200)
            c=1
        print("\n\n================== Mono/slit scans:")     
        if 'quick' in kwargs:
            energy(500)
            mono_energy_set(505)
            print('\n---------- Scanning slit 2B:\n') 
            flag=True
            while flag:
                scan_narrow_slit('2V',[0.25*c, -8, 8, 0.25])
                sleep(1)
                scan_narrow_slit('2H',[0.25, -6, 6, 0.25])
                sleep(1)
                m=BL.mda.lastFileNum()
                try:
                    V2=fit_mda(m-1,detDiode,1*c,'gauss')
                    plt.show()
                    H2=fit_mda(m  ,detDiode,1,'gauss')
                    plt.show()
                    list_position.append(V2)
                    list_position.append(H2)
                    #print(f"V2 = {round(V2,3)}")
                    #print(f"H2 = {round(H2,3)}")
                      
                    print('\nBeam position:')
                    print(f"   - vertical:   V1 = {round(V2,3)}")
                    print(f"   - horizontal: H1 = {round(H2,3)}")
                    v_steering=beamsteering_2B(V2)
                    print('\nM1 roll (RZ) steering:')
                    print('   - vertical: ~ '+v_steering)
                    print('   - horizontal: reset slit 2BH center to ',str(H2))
                    print('\n')         
                      
                    print('\n\nWARNING: 2V slit should always be centered at zero')
                    print('Increasing M1 roll moves the beam more positive on 2V:')
                    print('\t+0.05 pitch => ~ +0.04 slit position')
                    print('To steer M1, use:')
                    print("\tGet_Mirror(1);Move_M1('RZ',x)")
                except:
                    print('Unable to fit position.')
                if kwargs['sound']:playsound()
                if kwargs['interactive']: flag=interactive_fct('slit 2B')
                else: break
            print('\n---------- Scanning slit 1A:\n') 
            flag=True
            while flag:
                scan_narrow_slit('1V',[0.25, -4, 4, 0.1])
                sleep(1)
                scan_narrow_slit('1H',[0.25, -3, 3, 0.1])
                sleep(1)
                slits_set_BL()
                m=BL.mda.lastFileNum()
                try:
                    V1=fit_mda(m-1,detDiode,1,'gauss');plt.show()
                    H1=fit_mda(m  ,detDiode,1,'gauss');plt.show()
                    list_position.append(V1)
                    list_position.append(H1)
                    print('\nBeam position:')
                    print(f"   - vertical:   V1 = {round(V1,3)}")
                    print(f"   - horizontal: H1 = {round(H1,3)}")
                except:
                    print('Unable to fit position.')
                if kwargs['sound']:playsound()
                if kwargs['interactive']: flag=interactive_fct('slit 1A')
                else: break
            print('\nBeam center fit @ [2V,2H,1V,1H]:',list_position)
            print('\nTo update slit center using:     update_slit_dict()\n')
        else:    # longer full slit scans
            ID_SP_set(500)
            flag=True
            while flag:
                mono_energy_set(500)
                slits_set_BL()
                n=2 if 'extended_range' in kwargs  else 1
                scan_mono_vs_slit('2V',[0.25*c,-2*c*n,2*c*n,0.5],[475,515,2],comment='Mono/Slit - 2V')    #10/20 min for MEG/HEG
                scan_mono_vs_slit('2H',[0.25,-2*n,2*n,0.5],[475,515,2],comment='Mono/Slit - 2H')    #10min
                scan_mono_vs_slit('1V',[0.5,-0.75*n,0.75*n,0.25*c],[475,515,2],comment='Mono/Slit - 1V')    #10min
                scan_mono_vs_slit('1H',[0.5,-0.75*n,0.75*n,0.25],[475,515,2],comment='Mono/Slit - 1H')    #10min
                #interactive
                if kwargs['interactive']: flag=interactive_fct()
                else: break
                    
    ###### Check flux: 
    if 'flux' in kwargs['scanType']:
        if m3r_branch == 'c':
            diodeC('In')
        print("\n\n================== Check Flux:")
        flag=True
        while flag:
            check_flux(stay=True)
            exit_slit(50)
            energy(500)
            scanhv(475,525,1,comment='Mono Scan @ 500eV')
            try:
                m=BL.mda.lastFileNum()
                plot_mda(m,detDiode,flux=3);plt.show()
            except:
                print('Unable to fit position.')    
            ### interactive
            if kwargs['interactive']: flag=interactive_fct()
            else: break

    diagnostics_all_out()

    print("\n\n================== All done:")
    print("\nDon't forget to put back the user folder !!!!")
    
    ###### Plotting instructions and return first scan number 
    if FirstScan is not None:
        log_print(comment="\nFirstScan = "+str(FirstScan))
        print('StartOfTheWeek_plot("'+branch+'",'+str(FirstScan)+')' )
        print('Beam center fit @ [2V,2H,1V,1H]:',list_position)
    print('\nREMEMBER to update slit center using:     update_slit_dict()')
    
    if kwargs['sound']:
        playsound()
        
    return list_position
    
