from time import sleep
from os.path import dirname, join
import socket

from epics import caget, caput
from iexcode.instruments.userCalcs import userCalcOut_clear,userStringSeq_clear
from iexcode.instruments.scalers import Kappa_scaler
from iexcode.instruments.AD_utilities import AD_ROI_setup
from iexcode.instruments.scanRecord import ScanRecord
from iexcode.instruments.IEX_endstations import BL


"""
To do, get busy record in the mpa ioc, get user calcs in the mpa ioc


"""
##############################################################################################################
##############################                PVs             ##############################
##############################################################################################################
mpa_pv = "29iddMPA:"
mpa_busy = "29idcScienta:mybusy2"
userCalc_ioc = "29idTest:"
DAC_pv = '29iddau1:dau1:011:DAC'
HV_val_pv = "29idKappa:userCalcOut9.A"
HV_rbv_pv = "29idKappa:userCalcOut10.OVAL"
class MPA:
    """
    class for the MPA detector
    """

    def __init__(self,tth_pv, max_HV = 2990,ratio = 500,tth_dLLM=13.73,tth_dHLM=23.73):
        self.pv = mpa_pv
        self._reset_pv = mpa_pv+'C1O'
        self._det_pv = mpa_pv + "det1:"
        self._busy = mpa_busy
        self._DAC_pv = DAC_pv

        self.max_HV = max_HV
        self.ratio = ratio
        self.tth_dLLM = tth_dLLM
        self.tth_dHLM = tth_dHLM

        self._tth_pv = tth_pv
        self._reset_pv = mpa_pv+'C1O' #"writing a 1 resets"
        self._off_on_pv = mpa_pv+'C0O' #off/on = 0/1

        pass

    

    def HV_pvs(self):
        """
        returns the pvs for the userCalcout which hold the pvs for the mpa high voltage
        """
        val_pv = HV_val_pv
        rbv_pv =HV_rbv_pv
        return val_pv, rbv_pv

    ##############################################################################################################
    ##############################                HV Control              ##############################
    ##############################################################################################################

    def HV_set(self,volt,verbose=True):
        """
        sets the high voltage for the mpa

        Previously: MPA_HV_Set
        """
        val_pv, rbv_pv = self.HV_pvs()
        volt=min(volt,2990)

        caput(val_pv,volt,wait=True,timeout=18000)
        sleep(1)
        HV_rbv=caget(rbv_pv)

        if verbose:
            print("HV = "+str(HV_rbv)+" V")


    def HV_get(self):
        """
        sets the high voltage for the mpa
        """
        val_pv, rbv_pv = self.HV_pvs()

        return caget(rbv_pv)


    def HV_on(self):
        """
        turns the mpa high voltage on

        Previously: MPA_HV_ON
        """
        try:
            n_on=1
            tth_dial = caget(self._tth_pv+'.DRBV')

            if self.tth_dLLM <= tth_dial <= self.tth_dHLM:
                print('MPA OFF: detector in direct beam (-5 < tth for mcp < 5); move away before turning HV ON.')
            else:
                caput(self._reset_pv,1,wait=True,timeout=18000)
                caput(self._reset_pv,0,wait=True,timeout=18000)
                caput(self._off_on_pv,n_on,wait=True,timeout=18000)
                print("MPA - HV On")
        except:
            print('mpa not running')

    def HV_off(self):
        """
        turns the mpa high voltage off

        Previously: MPA_HV_OFF
        """
        try:
            n_off=0
            caput(self._off_on_pv,n_off,wait=True,timeout=18000)
            print("MPA - HV Off")
        except:
            print('MPA not running')


    def HV_reset(self):
        """
        resets the mpa high voltage

        Previously: MPA_HV_Reset
        """
        caput(self._reset_pv,1)
        print("MPA - Reset")
    
    
    def HV_scan(self,start=2400,stop=2990,step=10,**kwargs):
        """

        Previously: MPA_HV_scan
        """
        kwargs.setdefault('positioner_settling_time',1)
        Kappa_scaler(1)
        val_pv, rbv_pv = self._HV_pvs()
        BL.mda.fillin(val_pv, rbv_pv,start,stop,step,**kwargs)
        BL.mda.go(**kwargs)

    ##############################################################################################################
    ##############################                MCP Scripts              ##############################
    ##############################################################################################################
    def ROI_setup(self,ROI_num=1,xcenter=535,ycenter=539,xsize=50,ysize=50,binX=1,binY=1):  
        """
        usage:
            center of MCP, roiNum = 1 => MPA_ROI_SetUp(1,535,539,50,50)
            to set up all use: mpa_ROI_setup_all(xcenter,ycenter)
        """
        AD_ROI_setup('29iddMPA',ROI_num,xcenter,ycenter,xsize,ysize,binX,binY)
        self._ROI_stats(ROI_num)
    
    def ROI_setup_all(self,xcenter=535,ycenter=539):
        """
        setup up ROI 
            1 => size = 50 x 50
            2 => size = 100 x 100
            3 => size = 150 x 150
            4 => size = 200 x 200

        """
        self._ROI_setup(1,xcenter,ycenter,xsize=50,ysize=50)
        self._ROI_setup(2,xcenter,ycenter,xsize=100,ysize=100)
        self._ROI_setup(3,xcenter,ycenter,xsize=150,ysize=150)
        self._ROI_setup(4,xcenter,ycenter,xsize=200,ysize=200)
    
    
    def ROI_stats(self,ROI_num):
        """
        sequence to enable stats for mpa ROI
        """
        ROI_pv=self.pv+"ROI"+str(ROI_num)+':'
        stats_pv=self.pv+"Stats"+str(ROI_num)+':'
        caput(stats_pv+'NDArrayPort','ROI'+str(ROI_num))
        caput(stats_pv+'EnableCallbacks','Enable')
        caput(stats_pv+'ArrayCallbacks','Enable')
        caput(stats_pv+'ComputeStatistics','Yes')
        caput(stats_pv+'ComputeCentroid','Yes')
        caput(stats_pv+'ComputeProfiles','Yes')

    def _trigger_calcOut(self,**kwargs):
        """
        writes strSeq and calcOut for MPA this should go into the IOC

        **kwargs:
            ADplugin = 'TIFF1:' (default)
        """

        kwargs.setdefault("ADplugin","TIFF1:")
        kwargs.setdefault("save_image",False)

        ADplugin = self.pv + kwargs["ADplugin"]
        Proc1 = self.pv + "Proc1:"

        #All this should moved into the MPA IOC    
        Calcs_mda = ScanRecord(userCalc_ioc)
        
        desc = "MPA busy"
        n=9
        busy_CalcOut = userCalcOut_clear(Calcs_mda,n)
        caput(busy_CalcOut+".DESC",desc)
        caput(busy_CalcOut+".INPA",mpa_busy+" CP")
        caput(busy_CalcOut+".OOPT",'Transition to non-zero')
        caput(busy_CalcOut+".OUT",start_strSeq+".PROC PP")

        desc = "MPA start"
        n=1
        start_strSeq = userStringSeq_clear(Calcs_mda,n)
        caput(start_strSeq+".DESC",desc)
        caput(start_strSeq+".LNK1", Proc1+"ResetFilter PP")
        caput(start_strSeq+".STR1","Yes")

        desc = "MPA wait"
        n=10
        wait_CalcOut = userCalcOut_clear(Calcs_mda,n)
        caput(wait_CalcOut+".DESC",desc)
        caput(wait_CalcOut+".INPA",Proc1+"NumFilter_RBV")
        caput(wait_CalcOut+".INPB",Proc1+"NumFiltered_RBV")
        caput(wait_CalcOut+".OOPT",'Transition to non-zero')
        caput(wait_CalcOut+".OUT",done_strSeq+".PROC PP")

        desc = "MPA writeDone"
        n=1
        done_strSeq = userStringSeq_clear(Calcs_mda,n)
        caput(done_strSeq+".DESC","MPA writeDone")
        caput(done_strSeq+".LNK1", ADplugin+"WriteFile PP")
        caput(done_strSeq+".STR1","Write")
        caput(done_strSeq+".LNK1", mpa_busy+" PP")
        caput(done_strSeq+".STR1","Done")
            
    

    def self_trigger_callback(self,trigger,saveImg=False,**kwargs):
        """
        used for triggering the MPA in the scanRecord, reset Proc1 and waits for finish
        
        trigger: adds/removes trigger to mda (True/False)      
            
        **kwargs:
            save_image = False used for appropriate trigger of ROIs
                    = True also saves an image based on ADplugin type  
            ADplugin = 'TIFF1:' (default
        
        by default ADplugin = 29iddMPA:TIFF1: 
            can use 29iddMPA:HDF1:
        """

        kwargs.setdefault("ADplugin","TIFF1:")
        kwargs.setdefault("save_image",False)

        ADplugin = self.pv + kwargs['ADplugin']

        self._trigger_calcOut(**kwargs)
            
        if trigger==True:
            caput(ADplugin+"AutoResetFilter","Yes")
            if saveImg:
                caput(ADplugin+"AutoSave", "No")
                caput(ADplugin+"EnableCallbacks", "Enable")
            
        if trigger==False:
            caput(ADplugin+"AutoResetFilter","No")
            if saveImg:
                caput(ADplugin+"EnableCallbacks", "Disable")
                caput(ADplugin+"AutoSave", "Yes")

        return mpa_busy

    def _mpa_prefix(self,**kwargs):
        """
        """
        kwargs.setdefault("debug",False)
        
        fpath=join(dirname(dirname(BL.mda.filepath)),"mpa",'')
        nextMDA = BL.mda.fileNum
        prefix="mda"+str.zfill(str(nextMDA),4)+"_mpa"
        return prefix

    def _trigger(self,trigger, **kwargs):
        """
        Sets up / Cleans up the ScanRecord to trigger the MPA 

        trigger: adds/removes trigger to mda (True/False)
            
            
        **kwargs:
            save_image = False used for appropriate trigger of ROIs
                    = True also saves an image based on ADplugin type  
            detTrigNum = 2 (default)
            scan_dim = 1 (default)
            ADplugin = "TIFF1" / "HDF1"

        Previously: MPA_Trigger
        """
        kwargs.setdefault("detTrigNum",2)
        kwargs.setdefault("save_image",False)
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault("ADplugin","TIFF1:")
        
        mpa_busy = self._mpa_trigger_callback(trigger, **kwargs)
                        
        #adding the MPA to the scanRecord trigger
        if trigger == True:
            self._mpa_prefix(**kwargs)
            BL.mda.triggers_set(kwargs['scan_dim'],{2:mpa_busy})

        if trigger == False:
            BL.mda.triggers_set(kwargs['scan_dim'],{2:''})
        
   

    def _save_strSeq(self,**kwargs):
        """
        **kwargs:
            ADplugin: "TIFF1:"
        """
        kwargs.setdefault("ADplugin","TIFF1:")

        ADplugin = self.pv + kwargs["ADplugin"]

        desc = "MCP datamode"
        n=2
        strSeq_pv = userStringSeq_clear(BL.mda,n)
        caput(strSeq_pv+".DESC",desc)
        caput(strSeq_pv+".LNK1",self._det_pv+"Acquire CA NMS")
        caput(strSeq_pv+".STR1","Done")
        caput(strSeq_pv+".WAIT1","Wait")
        caput(strSeq_pv+".LNK2",self._det_pv+"RunTimeEnable PP NMS")
        caput(strSeq_pv+".STR2","1")
        caput(strSeq_pv+".WAIT2","Wait")
        caput(strSeq_pv+".LNK3",ADplugin+"EnableCallbacks PP NMS")
        caput(strSeq_pv+".STR3","1")

    def _freerun_strSeq(self):
        """
        """
        desc = "MCP freerun"
        n=1
        strSeq_pv = userStringSeq_clear(BL.mda,n)
        caput(strSeq_pv+".DESC",desc)
        caput(strSeq_pv+".LNK1",self._det_pv+"Acquire PP NMS")
        caput(strSeq_pv+".WAIT1","Wait")
        caput(strSeq_pv+".STR1","Done")
        caput(strSeq_pv+".LNK2","29iddMPA:TIFF1:EnableCallbacks PP NMS")
        caput(strSeq_pv+".STR2","0")
        caput(strSeq_pv+".WAIT2","Wait")
        caput(strSeq_pv+".LNK3",self._det_pv+"RunTimeEnable PP NMS")
        caput(strSeq_pv+".STR3","0")
        caput(strSeq_pv+".WAIT3","Wait")
        caput(strSeq_pv+".LNK4",self._det_pv+"Acquire PP NMS")
        caput(strSeq_pv+".STR4","Acquire")


    def _HV_sp_calcOut(self):
        """
        """

        desc = "MPA HV SP"
        n=9
        calcOut_pv = userCalcOut_clear(BL.mda,n)

        caput(calcOut_pv+".DESC",desc)
        caput(calcOut_pv+".A",0)
        caput(calcOut_pv+".B",self.ratio)
        caput(calcOut_pv+".C",self.max_HV)
        caput(calcOut_pv+".CALC$","A")
        caput(calcOut_pv+".OCAL$","MIN(A/B,C/B)")
        caput(calcOut_pv+".OOPT",1)        # On Change
        caput(calcOut_pv+".DOPT",1)        # Use 0CALC
        caput(calcOut_pv+".IVOA",0)        # Continue Normally
        caput(calcOut_pv+".OUT",self._DAC_pv+" PP NMS")

    def _HV_rbv_calcOut(self):
        """
        """
        desc = "MPA HV RBV"
        n=10
        calcOut_pv = userCalcOut_clear(BL.mda,n)

        caput(calcOut_pv+".DESC",desc)
        caput(calcOut_pv+".INPA",DAC_pv+' CP NMS')
        caput(calcOut_pv+".B",self.ratio)
        caput(calcOut_pv+".CALC$","A*B")

    def _interlock_mpa(self):
        """
        """
        desc = "MPA Interlock mpa"
        n=7
        calcOut_pv = userCalcOut_clear(BL.mda,n)

        caput(calcOut_pv+".DESC",desc)
        tth_pv = self._tth_pv
        caput(calcOut_pv+".INPA",tth_pv+".DRBV CP NMS")
        caput(calcOut_pv+".B",1)
        caput(calcOut_pv+".CALC$","ABS((("+str(self.tth_dLLM)+"<A && A<"+str(self.tth_dHLM)+") && (B>0))-1)")
        caput(calcOut_pv+".OCAL$",'A')
        caput(calcOut_pv+".OOPT",2)    # When zero
        caput(calcOut_pv+".DOPT",0)    # Use CALC
        caput(calcOut_pv+".IVOA",0)    # Continue Normally
        caput(calcOut_pv+".OUT",self._off_on_pv()+"PP NMS")

    
    def _interlock_DAC(self):
        """
        """
        desc = "MPA Interlock DAC"
        n=8
        calcOut_pv = userCalcOut_clear(BL.mda,n)

        caput(calcOut_pv+".DESC",desc)
        tth_pv = self._tth_pv
        caput(calcOut_pv+".INPA",tth_pv+".DRBV CP NMS")
        caput(calcOut_pv+".B",1)
        caput(calcOut_pv+".CALC$","ABS((("+str(self.tth_dLLM)+"<A && A<"+str(self.tth_dHLM)+") && (B>0))-1)")
        caput(calcOut_pv+".OCAL$",'A')
        caput(calcOut_pv+".OOPT",2)    # When zero
        caput(calcOut_pv+".DOPT",0)    # Use CALC
        caput(calcOut_pv+".IVOA",0)    # Continue Normally
        caput(calcOut_pv+".OUT",self._DAC_pv+"_Se PP NMS") 

def reset_mpa_HV():
    """
    resetting the SRS which controls the MPA HV
    Model: SRS PS350
    ip = "164.54.118.57"
    """
    ip = "164.54.118.57"
    # Open TCP connect to poet 1234 of GPIB-ETHERNET
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
    sock.settimeout(0.1)
    sock.connect((ip, 1234))

    # Set mode as CONTROLLER
    sock.send(b'++mode 1\n')

    # specify GPIB address of device being controlled
    addr = "14"
    str1="++addr"+addr+'\n'
    sock.send(bytes(str1,'utf-8'))

    # reset SRS PS350
    #sock.send("*RST\n")

    # turn the high voltage on, clearing any current or voltage trips
    sock.send(b"HVON\n")