"""
functions used for the mono
"""

import time
import numpy as np


from epics import caget,caput
from iexcode.instruments.IEX_endstations import *
from iexcode.instruments.utilities import print_warning_message,read_dict


##############################################################################################################
################################             mono             ##############################
##############################################################################################################
def mono_grating_names():
    """
    List of grating names 
    
    """
    ext = ['C',      'D',       'E',    'F',      'G',       'H',       'I',       "J",       "K",        "L"     ]
    GRT_names = ["MEG_Imp","HEG","MEG","Dummy",  "not used", "not used", "not used","not used","not used","not used"]
    GRT_names=GRT_names[0:2] # only use the first 3 slots
    return GRT_names

def mono_mirror_names():
    """
    List of grating names 
    
    """
    ext = ['C',      'D',       'E',    'F',      'G',       'H',       'I',       "J",       "K",        "L"     ]
    MIR_names = ["Au", "Silicon", "Carbon","not used","not used","not used","not used","not used","not used","not used"]
    MIR_names=MIR_names[0:2] # only use the first 3 slots
    return MIR_names

def mono_extensions():
    """
    extension used for the mono mirror and grating locations
    """
    ext = ['C',      'D',       'E',    'F',      'G',       'H',       'I',       "J",       "K",        "L"     ]
    return ext

def mono_energy_range(GRT=None):
    """
    returns hv_min,hv_max, the minimum and maxium energy achievable the the specified grating
    """
    if GRT == None:
        GRT = mono_grating_get()

    if GRT == 'HEG':
        hv_min=120
        hv_max=2200
    elif GRT == 'MEG':
        hv_min=120
        hv_max=3000
    return hv_min,hv_max

def mono_get_all(verbose=False):
    """
    gets hv, hv_sp,grating,mirror

    Previously: Mono_Optics
    """
    d = mono_get_all_extended(verbose=False)
    vals = {
       'hv':d['ENERGY_MON'],
       'hv_sp':d['ENERGY_SP'],
       'grating':d['GRT'],
       'mirror':d['MIR'],
    }
    if verbose:
        print(" hv: "+"%.2f" % vals['hv'], "eV, hv_sp: " % vals['hv_sp'])
        print(" grating  : "+vals['grating']," mirror  : "+vals['mirror'])
    return vals

def mono_grating_num():
    """returns the current grating num"""
    return caget('29idmonoGRT_TYPE_MON')

def mono_mirror_num():
    """returns the current mirror num"""
    return caget('29idmonoMIR_TYPE_MON')

def mono_pvs(grt_num=None,mir_num=None):
    """
    returns a dictionary with all the mono pvs
    """
    if grt_num is None:
        grt_num = mono_grating_num()
    if mir_num is None:
        mir_num = mono_mirror_num()

    ext=mono_extensions()

    d={
        'energy':"29idmono:ENERGY_MON",
        'energy_sp':"29idmono:ENERGY_SP",
        'grt_density':"29idmono:GRT_DENSITY",
        'grt_offset':'29idmonoGRT:P_OFFSETS.'+ext['grt_num'],
        'grt_b2':'29idmonoGRT:B2_CALC.'+ext['grt_num'],
        'grt_pos':'29idmonoGRT:X_DEF_POS.'+ext['grt_num'],
        'grt_LD':'29idmonoGRT:TYPE_CALC.'+ext['grt_num'],
        'grt_tun0':'29idmonoGRT:TUN0_CALC.'+ext['grt_num'],
        'grt_tun1':'29idmonoGRT:TUN1_CALC.'+ext['grt_num'],
        'grt_tun2':'29idmonoGRT:TUN2_CALC.'+ext['grt_num'],
        'grt_tun3':'29idmonoGRT:TUN3_CALC.'+ext['grt_num'],
        'grt_type_sp':"29idmonoGRT_TYPE_SP",
        'grt_move':"29idb:'gr:move",
        'grt_P_status':"29idmonoGRT:P_AXIS_STS",
        'mir_P_status':"29idmonoMIR:P_AXIS_STS",
        'grt_X_status':"29idmonoGRT:X_AXIS_STS",
        'mir_X_status':"29idmonoMIR:X_AXIS_STS",
        'mir_offset':'29idmonoMIR:P_OFFSETS.'+ext['mir_num'],
        'mir_pos':'29idmonoMIR:X_DEF_POS.'+ext['mir_num'],
        'cff':'29idmono:CC_MON',
        'arm':'29idmono:PARAMETER.G'

    }

def mono_get_all_extended(verbose=False):
    """
    gets the mono parameters for the current grating and mirror
    returns a dictionary with those values

    Previously: Mono_Optics
    """
    grt_num = mono_grating_num()
    mir_num = mono_mirror_num()
    grt_names = mono_grating_names()
    mir_names = mono_mirror_names()

    ext=mono_extensions()

    pvs = mono_pvs(grt_num=None,mir_num=None)
    vals={
        'grt_num':grt_num,
        'mir_num':mir_num,
        'grt':grt_names[grt_num],
        'mir':mir_names[mir_num],
        'energy':caget(pvs['energy']),
        'energy_sp':caget(pvs['energy_sp']),
        'grt_density':caget(pvs['energy_sp']),
        'grt_offset':caget(pvs['grt_offset']),
        'grt_b2':caget(pvs['grt_offset']),
        'grt_pos':caget(pvs['grt_pos']),
        'grt_LD':caget(pvs['grt_LD']),
        'grt_tun0':caget(pvs['grt_tun0']),
        'grt_tun1':caget(pvs['grt_tun1']),
        'grt_tun2':caget(pvs['grt_tun2']),
        'grt_tun3':caget(pvs['grt_tun3']),
        'grt_P_status':caget(pvs['grt_P_status']),
        'mir_P_status':caget(pvs['mir_P_status']),
        'grt_X_status':caget(pvs['grt_X_status']),
        'mir_X_status':caget(pvs['mir_X_status']),
        'mir_offset':caget(pvs['mir_offset']),
        'mir_pos':caget(pvs['mir_pos']),
        'CFF':caget('29idmono:CC_MON'),
        'arm':caget('29idmono:PARAMETER.G')
        }

    if verbose:
        for key in vals.keys(): 
            print(key,vals[key])

    return vals

def mono_grating_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['grt']

def mono_energy_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['energy']   

def mono_grating_density_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['grt_density']  

def mono_energy_set(hv_eV,verbose=True):
    """
    sets the mono energy

    Previously: SetMono
    """
    hv_min, hv_max = mono_energy_range()
    pv = mono_pvs()['energy_sp']
    if hv_min <= hv_eV <= hv_max: 
        caput(pv,hv_eV,wait=True,timeout=60)
    time.sleep(2.5)

    mono_status = mono_status_get()
    while True:
        if mono_status() > 1:
            mono_resest_pitch()
            caput(pv,hv_eV,wait=True,timeout=60)
            time.sleep(2.5)
        else:
            break
    if verbose:
        print("Mono set to",str(hv_eV),"eV")

def mono_scan_pvs():
    """
    returns the pvs to be used in scanning
    """
    val_pv = mono_pvs()['energy_sp']
    rbv_pv = mono_pvs()['energy']
    return val_pv, rbv_pv

def mono_status_get():
    """
    returns the status of the mirror (GRT * MIR)
        1 = ready


    Previously Mono_Status
    """
    pvs = mono_pvs()['energy']
    mir_P_status = caget(pvs['mir_P_status'])
    grt_P_status = caget(pvs['grt_P_status'])
    mir_X_status = caget(pvs['mir_P_status'])
    grt_X_status = caget(pvs['grt_P_status'])
    mirror_status = mir_P_status * mir_X_status
    grating_status = grt_P_status * grt_X_status
    return mirror_status*grating_status


def mono_grating_translate(grating,verbose=True):
    """
    Translate between the different grating positions

    Warning: this does not close the shutters use XX

    Previously: Move_GRT
    """
    current_grating=mono_grating_get()
    hv_eV = mono_energy_get()

    try:
        grating_state = mono_grating_names().index(grating)

    except:
        print_warning_message(grating+' not a valid grating')

    if current_grating != grating:
        pvs=mono_pvs()
        caput(pvs['grt_type_sp'],grating_state,wait=True,timeout=18000)        
        caput(pvs['grt_move'],1,wait=True,timeout=18000)
        while True:
            if mono_status_get() > 1:
                time.sleep(5)
            else:
                break
        mono_energy_set(hv_eV)

    else:
        if verbose:
            print("grating is already "+grating)

    if verbose:
        print("Mono Grating:",grating)


def mono_resest_pitch():
    """
    resets the MIR and GRT pitch interlocks after a following error or limit
    Previoulsy part of SetMono
    """
    caput("29idmonoMIR:P.STOP",1)
    time.sleep(1)
    caput("29idmonoGRT:P.STOP",1)
    time.sleep(1)
    print('Mono pitch was reset')

def mono_kill():
    """
    kills the mono pitch motors

    Previously: Kill_Mono
    """
    caput("29idmonoMIR:P_KILL_CMD.PROC",1)
    caput("29idmonoGRT:P_KILL_CMD.PROC",1)
#    caput("29idmonoGRT:X_KILL_CMD.PROC",1)
    time.sleep(5)
    caput("29idmono:STOP_CMD.PROC",1)
#    caput("29idmonoGRT:X_HOME_CMD.PROC",1)


def mono_stop():
    """
    presses the stop button on the mono motors 

    Previously: Stop_Mono 
    """
    caput("29idmono:STOP_CMD.PROC",1)
    time.sleep(5)

def mono_enable():
    """
    re-enable the mono motors 

    Previously: Enable_Mono
    """
    caput("29idmonoGRT:X_ENA_CMD.PROC",1)
    caput("29idmonoGRT:P_ENA_CMD.PROC",1)
    caput("29idmonoMIR:X_ENA_CMD.PROC",1)
    caput("29idmonoMIR:P_ENA_CMD.PROC",1)

def mono_limits_reset():
    """
    resest the mono low limit to 200 
    was changed in the pmac to the wrong value so we need to do this

    Previously: Reset_Mono_Limits
    """
    #    caput("29idmono_GRT_TYPE_SP.ONST", 'MEG_PA')
    #    caput("29idmono_GRT_TYPE_SP.TWST", 'HEG_JY')
    #    caput("29idmono_GRT_TYPE_SP.THST", 'MEG_JY')
    caput("29idmono:ENERGY_SP.DRVL",200)
    caput("29idmono:ENERGY_SP.LOW",200)
    caput("29idmono:ENERGY_SP.LOLO",200)
    caput("29idmono:ENERGY_SP.LOPR",200)
    print("Mono limits have been reset.")

def mono_cff_print():
    """
    print the cff tuning parameters and arm for the current grating

    Previously: Get_CFF
    """
    d = mono_get_all_extended
    cff = d['cff']
    tun0 = d['grt_tun0']
    tun1 = d['grt_tun1']
    tun2 = d['grt_tun2']
    tun3 = d['grt_tun3']
    arm =  d['arm']
    print(" cff : "+"%.4f" % cff , "          exit arm: "+"%.1f" % arm,"mm")
    print(" tun0 : "+"%.4e" % tun0 , "    tun1: "+"%.4e" % tun1)
    print(" tun2 : "+"%.4e" % tun2 , "    tun3: "+"%.4e" % tun3)

def mono_parameters_pv(grt_num=None,mir_num=None):
    """
    returns dictionary with mono_parameter for current grating/mirror
    """
    if grt_num is None:
        grt_num =  mono_grating_num()()
    if mir_num is None:
        mir_num = mono_mirror_num()()
   
    ext = mono_extensions()

    d={
        'mir_offset':"29idmonoMIR:P_OFFSETS."+ext[mir_num],
        'mir_pos':"29idmonoMIR:X_DEF_POS."+ext[mir_num],
        'grt_offset':"29idmonoGRT:P_OFFSETS."+ext[grt_num],
        'grt_pos':"29idmonoGRT:X_DEF_POS."+ext[grt_num],
        'grt_density':"29idmonoGRT:TYPE_CALC."+ext[grt_num],
        'grt_b2':"29idmonoGRT:B2_CALC."+ext[grt_num]
    }
    return d


def mono_parameters_get():
    """
    Gets the mono parameters and prints them for the History

    Previously: Mono_Parameters_Get
    """
    date=time.strftime("%Y%m%d",time.localtime())
    mirList = mono_mirror_names()
    grtList = mono_grating_names()
    pvList = mono_extensions()

    #MIR
    txt="MonoParm[\'"+date+"\']= {\n\t"
    for i in range(0,len(mirList)):
        d = mono_parameters_pv(mir_num=i)
        mir = mirList[i]
        offset = caget(d['mir_offset'])
        position = caget(d['mir_pos'])
        txt+="\'"+mir+"\':["+str(offset)+","+str(position)+"],"

    #GRT
    txt+="\n\t"
    for i in range(0,len(grtList)):
        d = mono_parameters_pv(grt_num=i)
        grt = grtList[i]
        offset = caget(d['grt_offset'])
        position = caget(d['grt_pos'])
        density = caget(d['grt_density'])
        b2 = caget(d['grt_b2'])
        txt+="\'"+grt+"\':["+str(offset)+","+str(position)+","+str(density)+","+str(b2)+"],"

    txt+="}"
    print(txt)

def mono_parameters_history_read(date):
    """
    Dictionary of Mono parameters used in Mono_Parameters(date) which sets the parameters
    #MonoParm['date']={'Au':[],Si:[],'Carbon':[],'MEG_Imp':[],'HEG_JY':[],'MEG_JY':[]}
    Mono_Parameters_Set(date) writes the parameters from the Dictionary
    
    Previously: Mono_Parameters_History
    """
    #MonoParm['date']={'Au':[],Si:[],'Carbon':[],'MEG_Imp':[],'HEG_JY':[],'MEG_JY':[]}

    MonoParm=read_dict('mono_parameters.txt')
    return MonoParm[date]


def mono_parameters_history_write(date):
    """

    Previously: Mono_Parameters_Set
    """
    hv_eV=caget("29idmono:ENERGY_SP")
    MonoParm=mono_parameters_history_read(date)
    pvList=['C','D','E']
    mirList=["Au","Si","Carbon"]
    for i in range(0,len(mirList)):
        mir=mirList[i]
        caput("29idmonoMIR:P_OFFSETS."+pvList[i],MonoParm[mir][0])
        caput("29idmonoMIR:X_DEF_POS."+pvList[i],MonoParm[mir][1])
    grtList=["MEG_Imp", "HEG_JY", "MEG_JY"]
    for i in range(0,len(grtList)):
        grt=grtList[i]
        caput("29idmonoGRT:P_OFFSETS."+pvList[i],MonoParm[grt][0])
        caput("29idmonoGRT:X_DEF_POS."+pvList[i],MonoParm[grt][1])
        caput("29idmonoGRT:TYPE_CALC."+pvList[i],MonoParm[grt][2])
        caput("29idmonoGRT:B2_CALC."+pvList[i],MonoParm[grt][3])
    time.sleep (1)
    mono_energy_set(hv_eV,verbose=True)



def mono_temperature_interlock():
    """
    userCalcOut to monitor the temperatures on the Mono if the temperature gets too high then it closes the main shutter
    #29idb:userCalcOut10.VAL =0 => OK; =1 => Too hot

    Previously: Mono_TemperatureInterlock
    """
    pv='29idb:userCalcOut10'
    caput(pv+'.DESC', 'Mono Temperature Interlock')
    caput(pv+'.INPA','29idmono:TC1_MON CP NMS')
    caput(pv+'.INPB','29idmono:TC2_MON CP NMS')
    caput(pv+'.INPC','29idmono:TC3_MON CP NMS')
    caput(pv+'.INPD','29idmono:TC4_MON CP NMS')
    caput(pv+'.INPE','29idmono:TC5_MON CP NMS')
    caput(pv+'.INPF','29idmono:TC6_MON CP NMS')
    
    caput(pv+'.H','31')
    
    caput(pv+'.CALC','A>H | B>H | C>H |D>H')
    caput(pv+'OOPT','Transition To Non-zero')
    caput(pv+'.OUT','PC:29ID:FES_CLOSE_REQUEST.VAL PP NMS')

    

def mono_scan_fillin(hv_start,hv_stop,hv_step,**kwargs):
    """
    fills in the scanRecord for scanning the mono

        puts the positioner in stay after scan

        **kwargs => scanRecord.fillin kwargs
        positioner_settling_time: increased because busy record is too fast => 0.2 (default)
    """
    kwargs.setdefault('positioner_settling_time',0.2)


    #Setting up the ScanRecord for Mono in Table mode
    val_pv, rbv_pv = mono_scan_pvs()
    BL.mda.fillin(val_pv,rbv_pv,hv_start,hv_stop,hv_step,**kwargs)

    #mono needs to stay and have a longer settling time
    BL.mda.positioner_after_scan("STAY")

    
def mono_scan_fillin_table(hv_array,**kwargs):
    """
    fills in the scanRecord for scanning the mono

    puts the positioner in stay after scan

    **kwargs => scanRecord.fillin kwargs
    positioner_settling_time: increased because busy record is too fast => 0.2 (default)

    """
    kwargs.setdefault('positioner_settling_time',0.2)

    #Setting up the ScanRecord for Mono in Table mode
    val_pv, rbv_pv = mono_scan_pvs()
    BL.mda.fillin.table(val_pv,rbv_pv,hv_array,**kwargs)

    #mono needs to stay and have a longer settling time
    BL.mda.positioner_settling_time(kwargs['positioner_settling_time'])
    BL.mda.positioner_after_scan("STAY")

def mono_scan_after(**kwargs):
    """
    resets mda after scanning the mono
    """
    after_scan_pv = BL.mda.default_after_scan_seq
    caput(after_scan_pv+".PROC",1)
    BL.mda.positioner_after_scan("PRIOR POS")


##############################################################################################################
################################             aligmnent and commissioning             ##############################
##############################################################################################################
def mono_motor_move(motor,value):
    """
    for pitch motor => MIR:P or GRT:P
    for translation motor => MIR:X or GRT:X

    Previously: Move_FMBMono
    """
    val_pv,rbv_pv = mono_motor_scan_pvs()
    caput(val_pv+".PREC",3)
    caput(val_pv,value,wait=True,timeout=18000)

def mono_motor_scan_pvs(motor):
    """
    returns the rbv and val for scanning 
    """
    val_pv = "29idmono"+motor+"_SP"
    rbv_pv = "29idmono"+motor+"_MON"
    return val_pv,rbv_pv

def mono_motor_scan_fillin(motor,start,stop,step,**kwargs):
    """
    for pitch motor => MIR:P or GRT:P
    for translation motor => MIR:X or GRT:X

    Previously: Scan_FMBMono
    """
    val_pv,rbv_pv = mono_motor_scan_pvs(motor)
    caput(val_pv+".PREC",3) # database sets .PREC==0.  We want more digits than that.

    BL.mda.fillin(val_pv,rbv_pv,start,stop,step,**kwargs)

def mono_zero_order(angle):
    """
    puts the beam in zero order => grating/mirror are parallel
    range ~1 - 5 degrees

    Previously: Mono_zero
    """
    angle=angle*1.0
    caput("29idmonoMIR:P_POS_SP",angle,wait=True,timeout=18000)
    caput("29idmonoMIR:P_MOVE_CMD.PROC",1,wait=True,timeout=18000)
    caput("29idmonoGRT:P_POS_SP",angle,wait=True,timeout=18000)
    caput("29idmonoGRT:P_MOVE_CMD.PROC",1,wait=True,timeout=18000)
    print("Mono set to zero order: MIR_pitch = "+str(angle)+", GRT_pitch = "+str(angle))

def mono_angles_set(alpha,beta): #JM modified to monitor the ready, moving sequentialy ended up in crash sometimes
    """
    Sets the mirror pitch (alpha) and grating pitch (beta) angles
    
    Previously: Mono_angle
    """
    alpha=alpha*1.0
    beta=beta*1.0
    #Putting Setpoints Go
    caput("29idmonoGRT:P_SP",alpha)
    caput("29idmonoMIR:P_SP",beta)
    ready=0
    while ready != 1:
        time.sleep(0.1)
        ready=caget('29idmono:ERDY_STS')
    print("Mono set to zero order: MIR_pitch = "+str(alpha)+", GRT_pitch = "+str(beta))

def mono_pink_beam():
    """
    moves grating and mirror out of beam
    x-rays will hit the pink beamstop

    Previously: Mono_pinkbeam
    """
    caput("29idmonoMIR:P_POS_SP",0,0)
    caput("29idmonoMIR:P_MOVE_CMD.PROC",0)
    time.sleep(3)
    caput("29idmonoGRT:P_POS_SP",0.0)
    caput("29idmonoGRT:P_MOVE_CMD.PROC",0)
    time.sleep(3)
    caput("29idmonoMIR:X_POS_SP",-52)
    caput("29idmonoMIR:X_MOVE_CMD.PROC",0)
    time.sleep(3)
    caput("29idmonoGRT:X_POS_SP",210)
    caput("29idmonoGRT:X_MOVE_CMD.PROC",0,wait=True,timeout=18000)

def mono_CC(): 
    """
    ##used for commissioning

    Previously: Mono_CC
    """
    caput("29idmonoMIR_TYPE_SP",3)
    caput("29idmonoMIR:X_DCPL_CALC.PROC",0)
    caput("29idmonoGRT_TYPE_SP", 10)
    caput("29idmonoGRT:X_DCPL_CALC.PROC",0)

    caput("29idmono:ENERGY_SP", 440)

def mono_grating_offset_set(val,grt_num=None):
    """
    sets the grating offset for the grating 

    Previously: Mono_Set_GRT0
    """
    pvs=mono_pvs(grt_num,None)
    caput(pvs['grt_offset'],val)
    mono_get_all_extended(verbose=True)

def mono_mirror_offset_set(val,mir_num=None):
    """
    sets the mirror offset for the grating 

    Previously: Mono_Set_MIR0
    """
    pvs=mono_pvs(None,mir_num)
    caput(pvs['mir_offset'],val)
    mono_get_all_extended(verbose=True)

def mono_grating_mirror_offsets_set(mirror_offset):
    """
    changes the mirror and grating offsets to  maintain parallelism
    After determining parallelism (zero order a=b)
    Get the correct take-off angle delta, equivalent to scan ExitSlit_Vcenter
    
    Previously: Mono_Set_MIR0_GRT0,MIR_GRT_Offset,Mono_Set_MIR0_GRT0_all
    """  
    current_vals = mono_get_all_extended(verbose=False)
    energy_sp = current_vals['energy_sp']
    mirror_offset_old = current_vals['mir_offset']
    
    #get current grating offsets
    grating_offsets=[]
    for grt_num in range(0,3):
        grating_offsets.append(mono_parameters_pv(grt_num=grt_num,mir_num=None)['grt_offset'])

    #calc mirror_offset - grating_grating    
    mir_grt_offsets = mirror_offset_old-np.array(grating_offsets)    # list of MIR-GRT for all 3 gratings

    #set mirror_offset
    mono_mirror_offset_set(mirror_offset)

    #set grating_offsets
    for grt_num in range(0,3):
        mono_grating_offset_set(mirror_offset - mir_grt_offsets[grt_num],grt_num)

    time.sleep (1)
    mono_energy_set(energy_sp)
    mono_get_all_extended(verbose=True)


def mono_grating_b2_set(val,grt_num=None):
    """
    sets the grating offset for the grating 
    Previously: Mono_Set_b2
    """
    hv = mono_energy_get()
    pvs = mono_pvs(grt_num,None)
    caput(pvs['grt_b2t'],val)
    time.sleep(1)
    
    mono_energy_set(hv)
    mono_get_all_extended(verbose=True)

def mono_cff_set(val,tune_num):
    """
    sets the tuning parameters for the cff calculation 
    tune_num = order

    Previously: Mono_Set_cff
    """
    hv = mono_energy_get()
    pvs = mono_pvs()

    #get current info
    current_vals = mono_get_all_extended(verbose=False)
    grt_pitch = current_vals['grt_pitch']
    mir_pitch = current_vals['mir_pitch']
    print ('Pitch grating, mirror: ',grt_pitch,mir_pitch)
    
    #set the tuning coeffient
    tun='tun'+str(tune_num)
    caput(pvs[tun],val)
    time.sleep(1)

    #set then energy
    mono_energy_set(hv)

    #print differences in pitch
    mono_get_all_extended(verbose=False)
    grt_dif = grt_pitch-current_vals['grt_pitch']
    mir_dif = mir_pitch-current_vals['mir_pitch']

    print('Pitch grating, mirror: ',current_vals['grt_pitch'],current_vals['mir_pitch'])
    print('Differences          : ',grt_dif,mir_dif)

def mono_arm_set(distance_mm):
    """
    sets the exit arm for the grating 

    Previously: Mono_Set_ExitArm
    """
    hv = mono_energy_get()
    pvs = mono_pvs()

    #get current info
    current_vals = mono_get_all_extended(verbose=False)
    grt_pitch = current_vals['grt_pitch']
    mir_pitch = current_vals['mir_pitch']
    print ('Pitch grating, mirror: ',grt_pitch,mir_pitch)
    
    #set the exit arm
    caput("29idmono:PARAMETER.G",distance_mm)
    time.sleep(1)

    #set then energy
    mono_energy_set(hv)

    #print differences in pitch
    mono_get_all_extended(verbose=False)
    grt_dif = grt_pitch-current_vals['grt_pitch']
    mir_dif = mir_pitch-current_vals['mir_pitch']

    print('Pitch grating, mirror: ',current_vals['grt_pitch'],current_vals['mir_pitch'])
    print('Differences          : ',grt_dif,mir_dif)