"""
functions related to using x-rays

"""
import numpy as np
from time import sleep

from epics import caget,caput
from iexcode.instruments.IEX_endstations import BL

from iexcode.instruments.IEX_VPU import *
from iexcode.instruments.VLS_PGM import *
from iexcode.instruments.slits import *
from iexcode.instruments.shutters import *
from iexcode.instruments.gate_valves import *
from iexcode.instruments.diagnostics import *
from iexcode.instruments.m3r import *
from iexcode.instruments.logfile import *
from iexcode.instruments.utilities import print_warning_message,make_table, take_closest_value
from iexcode.instruments.current_amplifiers import ca_average
from iexcode.instruments.beamline import endstation_cams_enable

from iexcode.instruments.MPA import MPA


##############################################################################################################
##############################            resets and detector lists         ##############################
##############################################################################################################
def test_function():
    print('hello world')
    
def xrays_reset():
    mono_limits_reset()
    xray_motor_encoder_sync()


def xrays_detector_list():
    """
    defualt detectors for the APRES endstation
    used by: 
         CA_Live_StrSeq()
        Detector_Triggers_StrSeq()
        BeforeScan_StrSeq() => puts everybody in passive
        CA_Average()
    WARNING: can't have more than 5 otherwise CA_Live_StrSeq gets angry.

    Previously: part of Detector_List
    """
    ca_list=[["b",4],["b",13]]

    return ca_list


def xrays_detector_dictionary():
    """
    returns a dictionary of the default detectors
    
    Previously: Detector_Default
    """
    det_dict={}
    sr_main_shutter={
        1:"S:SRcurrentAI.VAL",
        2:"EPS:29:ID:SS1:POSITION"
        }
    bl_energy={
        3:"29idmono:ENERGY_MON",
        4:"ID29:EnergySet.VAL",
        5:"ID29:Energy.VAL"
        }
    bl_keithleys={
        6:"29idb:ca1:read",
        7:"29idb:ca2:read",
        8:"29idb:ca3:read",
        9:"29idb:ca4:read",
        10:"29idb:ca5:read",
        11:"29idb:ca9:read",
        12:"29idb:ca12:read",
        13:"29idb:ca13:read"
        }
    det_dict.update(sr_main_shutter)
    det_dict.update(bl_energy)
    det_dict.update(bl_keithleys)
    return det_dict

##############################################################################################################
##############################            get beamline info         ##############################
##############################################################################################################

def xrays_get_all(verbose=False):
    """
    gets info about current beamline setting
    returns a dictionary

    """
    vals={}
    print("\n===========================================================")
    vals.update(ID_get_all(verbose=True))
    vals.update(mono_get_all(verbose=True))
    print("-----------------------------------------------------------")
    vals.update(slits_get_all(verbose=True))
    vals.update(FMB_mirror_get(0,verbose=True))
    vals.update(FMB_mirror_get(1,verbose=True))
    vals.update(FMB_mirror_get(3,verbose=True))
    print("-----------------------------------------------------------")
    vals.update({BL.endstation:BL.Motors.mprint()})
    print(BL.endstation+" = ",BL.Motors.mprint())
    
    print("===========================================================")
    return vals

def xrays_log_entries(**kwargs):
    """
    Writes CSV file for the MDA scans with the following default parameters:
    FilePath='/home/beams/29IDUSER/Documents/User_Folders/UserName/'
    Filename is set by logname_set(FileName=None), default:YYYYMMDD_log.txt, where YYYYMMDD is when folders were set
        (FileName=None,FileSuffix=None) is they are specified then it will use those values instead

    comment="Additional comments, last column"

    Previously: scanlog
    """
    #default parameters
    kwargs.setdefault('comment','')

    ioc = BL.ioc

    #scan Num
    scan = BL.mda.prefix(ioc)+str(BL.mda.fileNum(ioc))
    entry_list=["scan"]
    pv_list=[scan]
    format_list=["s" ]
    #scan info
    drive = caget(ioc+"scan1.P1PV")
    start = caget(ioc+"scan1.P1SP")
    stop  = caget(ioc+"scan1.P1EP")
    step  = caget(ioc+"scan1.P1SI")
    entry_list.append('drive','start','stop' ,'step')
    pv_list.append(drive,start,stop ,step)
    format_list.append("s",".3f",".3f",".3f")

    #endstation info
    if BL.endstation_name == 'ARPES':
        endstation_entry, endstation_pv, endstation_format = BL.log_entries(**kwargs)
        entry_list.append(endstation_entry)
        pv_list.append(endstation_pv)
        format_list.append(endstation_format)
    elif BL.endstation_name == 'Kappa':
        endstation_entry, endstation_pv, endstation_format = BL.log_entries(**kwargs)
        entry_list.append(endstation_entry[:-7])
        pv_list.append(endstation_pv[:-7])
        format_list.append(endstation_format[:-7])

    #beamline info
    ID_mode = ID_mode_get(verbose=False) if BL.xrays else "no_xrays"
    ID_QP_ratio = ID_QP_ratio_get (verbose=False) if BL.xrays else "no_xrays"
    ID_sp  = round(ID_SP_get_eV()) if BL.xrays else 0
    ID_rbv = round(ID_rbv_get_eV,4) if BL.xrays else 0
    hv = mono_energy_get() if BL.xrays else 0
    grt = mono_grating_get() if BL.xrays else 0
    slit_size = slit_get()[0] if BL.xrays else 0
    entry_list.append("hv","exit_slit","GRT", "ID_SP", "ID_RBV", "ID_Mode", "ID_QP")
    pv_list.append(hv,slit_size,grt, ID_sp, ID_rbv,ID_mode, ID_QP_ratio)
    format_list.append(".2f",".0f","s",".1f",".1f","s", ".0f")

    #endstation info 2: Fanny can I change to order to get rid of this complication?
    if BL.endstation_name == 'Kappa':
        endstation_entry, endstation_pv, endstation_format = BL.log_entries(**kwargs)
        entry_list.append(endstation_entry[-7:])
        pv_list.append(endstation_pv[-7:])
        format_list.append(endstation_format[-7:])

    #timestamp and comments
    t = time.strftime("%D-%H:%M:%S")
    comment = kwargs["comment"]
    entry_list.append('time','comment')
    pv_list.append(t,comment)
    format_list.append("s","s")

    try:
        logfile_update(BL.endstation,BL.ioc,entry_list,pv_list,format_list)
    except:
        print("scanlog did not write to file, check for errors.")
        
##############################################################################################################
##############################            beamline encoders         ##############################
##############################################################################################################

def xray_motor_encoder_sync():
    ioc = "29idb:"
    for motor in [13,14,15,16,26,27]:
        pv = ioc+"m"+str(motor)+".SYNC"
        caput(pv,1)
        print("synch encoder: "+pv)

##############################################################################################################
##############################               energy : ID + mono           ##############################
##############################################################################################################
def energy_get_all(verbose=True):
    """
    returns ID_mode,ID_QP_ratio,ID_sp,ID_rbv,hv,grt 

    """
    d = ID_get_all(verbose=False)
    d.update(mono_get_all(verbose=False))
    keylist = ['ID_Mode', 'ID_QP_ratio', 'ID_SP', 'ID_RBV','hv','grating']
    values = []
    for key in keylist:
        values.append(d[key])
        if verbose:
            print(key,d[key])

    return tuple(values)
    
def energy_get():
    hv = getE()
    return hv

def energy(hv_eV,slit_coeff=1,m3r=True,verbose=True):
    """
    sets the ID for optimal flux based on calibration
    sets the mono
    sets the beamline aperture/slits

    slit_coeff < 1 is used to kill flux by closing the aperature after the mono (slit-2B)
    
    m3r => if True optimizes the mirror for the d-branch only
    Previously: Set_BL, energy
    """
    if BL.xrays:
        if hv_eV != energy_range_check(hv_eV):
            message_string = 'request photon energy '+str(hv_eV)+' not with the allowed range'
            message_string = '\n closest allowed energy is '+str(energy_range_check(hv_eV))
            print_warning_message(message_string)


        ID_energy_set(hv_eV,verbose=verbose)
        mono_energy_set(hv_eV,verbose=verbose)
        slits_set_BL(c_2B=slit_coeff,c_1A=1,verbose=verbose)

        if m3r == True:
            if BL.branch() == 'd':
                print('\nalign_m3r()')
                try:
                    m3r_align()
                    sleep(1)
                    if m3r_RY_pos_sp_get() == m3r_RY_pos_sp():
                        m3r_align()
                except:
                    print('Unable to align; check camera settings.')
    else:
        message_string = 'BL.xrays = False, energy is not set'
        print_warning_message(message_string)

def energy_range_min_max():
    """
    returns the min,max energies for the current grating/ID setting
    """
    ID_min, ID_max = ID_energy_range()
    mono_min, mono_max = mono_energy_range()
    hv_max = np.min(ID_max, mono_max)
    hv_min = np.max(ID_min, mono_min)
    return hv_min,hv_max

def energy_range_check(hv):
    """
    checks if the energy is within the allowed range for both the mono
    and the ID. If outside the ranges set it to the closest allowed value

    Previously: SetRange()
    """
    hv_min,hv_max = energy_range_min_max()
    hv = np.min(hv_max,hv)
    hv = np.max(hv_min,hv)
    return hv

def mvid(val):
    """
    Sets the ID absolute set point (not optimized , mono & apertures stay fixed).
    
    to go to most flux use energy
    """
    ID_SP_set()

def scan_ID(ID_sp_start,ID_sp_stop,ID_sp_step,**kwargs):
    """
    scan the ID set point
    """
    val_pv,rbv_pv = ID_scan_pvs()
    BL.mda.fillin(val_pv,rbv_pv,ID_sp_start,ID_sp_stop,ID_sp_step,**kwargs)


def mvmono(val):
    """
    Sets the mono energy (insertion device & apertures stay fixed).
    """
    mono_energy_set(val)

def getE():
    """ 
    Returns current mono set point.
    """
    return mono_energy_get()

def switch_gratings(grating):
    """
    used to switch between gratings
        grating = 'HEG' / 'MEG'

        check current grating 
        closes the main shutter 
        sets the mono energy 

    Previously: Switch_Grating
    """
    current_grt = mono_grating_get()
    if grating != current_grt:
        shutter_open = main_shutter_status()
        main_shutter_close()
        print("Switching grating, please wait...")
        mono_grating_translate(grating,quiet=True)
        slits_set_BL()

        if shutter_open:
            main_shutter_open()
    
        print(dateandtime()," grating: "+grating)
    else: 
        print("grating: "+grating)

def mvgrating(grating):
    """
    Change mono grating: 
        HEG = high resolution, low flux
        MEG = medium resolution, high flux
    """
    switch_gratings(grating)

def polarization(ID_mode):
    """
    Change beam polarization: ID_mode = 'H', 'V', 'RCP' or 'LCP'
    """
    ID_switch_mode(ID_mode)

def polarization_get():
    """
    gets the current polarization
    """
    ID_mode_get()

def scanhv(start,stop,step,average_pnts=1,**kwargs):
    """
    scans the mono at the current ID value

    """
    ca_average(average_pnts)

    mono_scan_fillin(BL.mda,start,stop,step,**kwargs)

    mono_energy_set(start)
    BL.mda.go(**kwargs)

    mono_scan_after(BL.mda)


def scanE(start,stop,step,ID_offset=0,mesh='stay',average_pnts=1,scan_dim=1,**kwargs):
    """ 
    sets the ID and then scans the mono
    
    hv_start,hv_stop,hv_step: are the mono start, stop, step energies
    ID_offset: used to position the ID
        None => the ID does not move
        otherwise ID_energy_set = (stop-start)/2 + ID_offset
    mesh: used to specifiy mesh status (for normalization)
        => mesh='y': drops in the mesh for normalization (D branch only); retracts it at the end.
        => mesh='stay': to leave the mesh in after the scan is finished
        => mesh='n': no mesh.
    **kwargs
        scan_dim
        positioner_settling_time = 0.2
    """
    mda = BL.mda
    ca_average(average_pnts)

    if ID_offset != None:
        ID_energy_set = (start+stop)/2.0 + ID_offset
        
    slits_set_BL()

    mono_scan_fillin(mda,scan_dim,start,stop,step,**kwargs)
    
    mono_energy_set(start)
    BL.mda.go(scan_dim)
    
    mono_scan_after(mda,scan_dim)


def scanXAS(ID_eV,start_stop_step_lists,**kwargs):
    """
    Sets the beamline to ID_eV and then scans the mono for XAS scans

    start_stop_step_lists is a list of lists for the different scan ranges
        start_stop_step_lists = [[start1,stop1,step1], [start1,stop1,step1],...]
        Note duplicates are removed and the resulting array is sorted in ascending order

    **kwargs:
        scan_dim = 1 (default)
        average_pnts: if using Keithlys this set the averaging; None for no averaging
        m3r: if True the mirror is optimized before the start of the scan
        mcp
        execute: True/False to start the scan => True (default)
    
    Normalization:
        - If in the D-branch the Mesh is put in but is not removed
        - If in the C-branch we use the slit blades for normalization (ca13)
    """
    kwargs.setdefault("scan_dim",1)
    kwargs.setdefault("average_pnts",1)
    kwargs.setdefault("m3r",True)
    kwargs.setdefault("mcp",True)    
    kwargs.setdefault('execute',True)   

    scan_dim=kwargs['scan_dim']
    
    #Setting up the ScanRecord for Mono in Table mode
    hv_array = make_table(start_stop_step_lists)
    mono_scan_fillin_table(BL.mda,scan_dim,hv_array,**kwargs)

    #Averaging and Normalization
    ca_average(kwargs['average_pnts'])
    if BL.branch=="d":
        meshD("In")
     
    #Setting the beamline energy
    energy(ID_eV,m3r=kwargs["m3r"])

    #mpa
    if BL.branch == "d" and kwargs["mcp"]:
        mpa_HV_on()

    #Scanning
    if kwargs['execute']:
        mono_energy_set(hv_array[0])
        BL.mda.go(scan_dim)
    
        #Setting everything back
        mono_energy_set(ID_eV)
        mono_scan_after(BL.mda,scan_dim)
        BL.mda.table_reset_after(scan_dim)

        if BL.branch == "d": 
            if kwargs["mcp"]: 
                mpa_HV_off()
            print("WARNING: Mesh"+BL.branch+" is still In")        

def scanXAS_BL(start_stop_step_lists,**kwargs):
    """
    scans the mono and the ID for XAS scans
        Note: this is slow, but required in if the c-branch

    start_stop_step_lists is a list of lists for the different scan ranges
        start_stop_step_lists = [[start1,stop1,step1], [start1,stop1,step1],...]
        Note duplicates are removed and the resulting array is sorted in ascending order

    **kwargs:
        scan_dim = 1 (default)
        average_pnts: if using Keithlys this set the averaging; None for no averaging
        m3r: if True the mirror is optimized before the start of the scan
        mcp
        execute: True/False to start the scan => True (default)
    
    Normalization:
        - If in the D-branch the Mesh is put in but is not removed
        - If in the C-branch we use the slit blades for normalization (ca13)
    """
    kwargs.setdefault("scan_dim",1)
    kwargs.setdefault("average_pnts",1)
    kwargs.setdefault("m3r",True)
    kwargs.setdefault("mcp",True)    
    kwargs.setdefault('execute',True)

    scan_dim=kwargs['scan_dim']
        
    #Setting up the ScanRecord for Mono and ID in Table mode
    mono_array,ID_array = BL_energy_tables(start_stop_step_lists)

    kwargs.update('positioner_num',1)
    mono_scan_fillin_table(BL.mda,scan_dim,mono_array,**kwargs)

    kwargs.update('positioner_num',2)
    ID_scan_fillin_table(BL.mda,scan_dim,ID_array,**kwargs)
    
    #Averaging and Normalization
    ca_average(kwargs['average_pnts'])
    if BL.branch=="d":
        meshD("In")
     
    #Setting the beamline energy
    slits_set_BL(mono_array[0],m3r=kwargs["m3r"])

    #mpa
    if BL.branch == "d" and kwargs["mcp"]:
        mpa_HV_on()
        
    #Scanning
    if kwargs['execute']:
        mono_energy_set(mono_array[0])
        ID_SP_set(ID_array[0])
        BL.mda.go(scan_dim)

        #Setting everything back
        mono_energy_set(mono_array[0])
        mono_scan_after(BL.mda,scan_dim)
        BL.mda.table_reset_after(scan_dim)

        if BL.branch == "d": 
            if kwargs["mcp"]: 
                mpa_HV_off()
            print("WARNING: Mesh"+BL.branch+" is still In")  

def BL_energy_tables(start_stop_step_lists):
    """
    returns mono_array and ID_array for BL energy scans
    *energies:
        start,stop,step
        ListofLists, e.g.{[400,420,1],[420,425,0.25]...}

    Previously: Tables_BLenergy
    """
    mono_array = make_table(start_stop_step_lists)
    ID_array=np.array([])
    ID_mode,ID_QP_ratio,ID_sp,ID_rbv,hv,grt = energy_get_all(verbose=False)
    for hv_eV in mono_array:
        ID_array=np.append(ID_array,ID_calc_SP(grt,ID_mode,hv_eV))
    return mono_array,ID_array


##############################################################################################################
###########################               branch shutter, valves, safe_state                    ######################
##############################################################################################################
def get_branch(verbose=True):
    """
    gets the branch based on the position of m3r
    """
    branch = m3r_branch(verbose)
    return branch


def open_branch(branch=None,valve=False):
    """
        Opens the branch shutter, if branch is none, then uses the position of m3r to determine branch
        
        Previously: Check_BranchShutter
    """
    branch = get_branch(verbose=False)
    
    if branch_shutter_status(branch):
        print(dateandtime(), +branch+" shutter already open..." ,)
    else:
        while True:
            if branch_shutter_status(branch) == False :
                print(dateandtime(), "Opening "+branch+" shutter ..." ,)
                branch_shutter_open(branch,verbose=False)
                sleep(30)
            else:
                print(dateandtime(), +branch+" shutter is now open..." ,)
                break
    if valve:
        GVs = branch_valves()
        for GV in GVs:
            valve_open(GV)

def close_shutter(branch=None):
    """
        Closes the branch shutter, if branch is none, then uses the position of m3r to determine branch
        
        Previously: Close_CShutter,Close_DShutter
    """
    if branch is None:
        branch = get_branch()
    caput("PC:29ID:S"+branch+"S_CLOSE_REQUEST.VAL",1,wait=True,timeout=18000)
    print("Closing "+branch+"-Shutter...")

def switch_branch(branch, force=False, shutter=True,scan_reset=True,enable_cams=True):
    """Switch beam into which = \"c\" or \"d\" branch (by retracting/inserting deflecting mirror)
        Optionnal keyword argument:
            force: change m3r even if the beamline think it is already there (default=False)
            shutter: open the branch shutter, does not open the valve (default=False) 
            reset: resets the scanRecord and the global BL for the selected branch
            enable_cams: to turn on/off branch camera  

    Previously: Switch_Branch
    """
    branch = branch.lower()
    if branch in ["c","d","e"]:

    # Check current branch:
        if get_branch() == branch and not force:
            m3r_position = FMB_mirror_get(3,verbose=False)
            if np.sum(np.array(m3r_position)):
                print("\nMirror homed...")
                print("...Try again using: Switch_Branch(which, forced=True)")
            else:
                print("\nWell, looks like the beam is already in this branch...")
                print("...if you want to force the mirror motion, use the argument: forced=True")
        else:
            # Close both shutters:
            print("\n")
            print("Switching branch, please wait...")
            if shutter:
                branch_shutter_close('c')
                branch_shutter_close('d')
    
            #Move M3R:
            m3r_switch_branch(branch)
            print(dateandtime()," -  In "+branch+"-Branch")
            sleep(2)

        # Open branch shutters:
            if shutter:
                open_branch(branch,valve=False)

            if scan_reset:
                BL.mda.reset()

            if enable_cams:
                endstation_cams_enable(branch)

    else:
        print_warning_message(branch+' is not a valid branch selection')
#             if not nocam:

##############################################################################################################
###########################                      mirrors                    ######################
##############################################################################################################
def get_mirrors():
    for mirror_num in [0,1,3]:
        FMB_mirror_get(mirror_num,verbose=True)


##############################################################################################################
###########################                  slits and resolution                    ######################
##############################################################################################################
def slit(size):
    """
    sets the exit slit based on BL.endstation

        ARPES = 0 < x < 300  um
        Kappa  = 0 < x < 1000 um
    """
    branch = BL.branch
    if branch == "c":
        slit3C_set(size, quiet=False)
    elif branch == "d":
        slit3D_set(size, quiet=False)


def slit_get(verbose=True):
    """
    sets the exit slit based on BL.endstation

        ARPES = 0 < x < 300  um
        Kappa  = 0 < x < 1000 um
    """
    branch = BL.branch
    if branch == "c":
        slit_size = slit3C_get(verbose=False)
        slit_center = np.nan
    elif branch == "d":
        slit_size,slit_center = slit3D_get(verbose=False)

    if verbose:
        message = BL.branch+' exit slit: '+str(slit_size)
        message += ", @ "+str(slit_center) if BL.branch is not 'c' else ''
        print(message)    
    return slit_size,slit_center





